<?php

namespace BotKit\Common;

// Команды бота

use BotKit\Common\Message;
use BotKit\Attachments\ImageAttachment;

use BotKit\Enums\Platform;
use BotKit\Enums\State;
use BotKit\Enums\GroupSelect;
use BotKit\Enums\AccountType;
use BotKit\Enums\DateSelect;
use BotKit\Enums\CacheType;

use BotKit\Keyboards\TermsOfServiceKeyboard;
use BotKit\Keyboards\SelectUserTypeKeyboard;
use BotKit\Keyboards\HubKeyboard;
use BotKit\Keyboards\SelectGroupCourseKeyboard;
use BotKit\Keyboards\AskEnterCredentialsKeyboard;
use BotKit\Keyboards\DateSelectKeyboard;

use BotKit\Drivers\Driver;

use BotKit\Entities;

class Commands {

    #region Темы

    // Темы необходимы для настройки отрисовки таблиц - например расписания
    // Каждая тема ОБЯЗАНА иметь следующие ключи:
    // line_spacing - расстояние в пикселях между двумя строками текста
    // padding - сколько пикселей добавлять яйчейкам таблицы (см. css box model)
    // colors - массив цветов
    // в colors тоже есть обязательные ключи
    //   background - цвет заднего фона
    //   title_color - цвет текста заголовка
    //   body_bg_even - фон чётных строк таблицы
    //   body_bg_odd - фон нечётных строк таблицы
    //   body_fg - цвет текста в строках таблиц
    // title_font_size - размер шрифта (pt) заголовка
    // body_font_size - размер шрифта (pt) в таблице
    // body_line_constraints - ограничения на символы в строках таблицы
    // title_line_size - ограничения на символы в заголовке таблицы
    // font - путь к файлу шрифта

    // В тему разрешается добавлять и другие ключи, как, например в теме оценок
    
    private static array $theme_schedule =
    [
		"line_spacing"=> 5,
		"padding"=> 10,
		"colors"=> [
			"background"=> [220,220,220],
			"title_color"=> [30, 30, 30],
			"body_bg_even"=> [180, 180, 170],
			"body_bg_odd" => [210, 210, 200],
			"body_fg" => [40, 40, 40],
            "body_frame" => [25, 25, 25]
		],
		"title_font_size"=> 30,
		"body_font_size"=> 20,
        "body_line_constraints" => [0, 40, 25],
        "title_line_constraints" => 35,
        "font" => rootdir.'/resources/Lato-Regular.ttf',
        "body_line_height" => 25,
        "title_line_height" => 35
	];
    #endregion

    #region Callback

    // Отправка условий использования
    public static function showTermsOfService($e, $u, $driver) {
        $driver->editMessage(
            $e->getMessageID(),
            Message::create("[Условия использования]")
        );
    }

    // Выбор группы
    public static function handleGroupSelection($e, $u, $driver) {
        $params = $e->getParams();
        $intent = GroupSelect::from($params['intent']);
        $msg_id = $e->getMessageID();
        $em     = Database::getEM();

        // Получение выбранной группы по ID
        $group_id = $params['group_id'];
        $group = $em->find(Entities\CollegeGroup::class, $group_id);
        if ($group == null) {
            $driver->sendMessage(
                $msg_id,
                Message::create("Не найдена группа с id ".$group_id));
            return;
        }
        
        switch ($intent) {
            case GroupSelect::Register:
                // Группа выбрана, зарегистрирован студент
                $student = $em->getRepository(Entities\Student::class)
                    ->findByUser($u);
                $student->setGroup($group);
                
                $driver->editMessage(
                    $e->getMessageID(),
                    Message::create("Хочешь ввести логин и пароль для просмотра оценок?")
                        ->withKeyboard(new AskEnterCredentialsKeyboard())
                );
                break;

            default:
                $driver->sendMessage(
                    $u,
                    Message::create("Выбрана неизвестная группа. ID=".$group_id));
                break;
        }
    }

    // Выбор хочу/не хочу
    public static function enterCredentials($e, $u, $driver) {
        $params = $e->getParams();
        if ($params['enters'] == 0) {
            // Пользователь не захотел вводить пароль и логин
            $driver->editMessage(
                $e->getMessageID(),
                Message::create("Регистрация завершена!")
                    ->withKeyboard(new HubKeyboard())
            );
            $u->setState(State::Hub);
        } else {
            // Пользователь захотел вводить пароль и логин
            $driver->sendMessage($u, Message::create("Введи логин"));
            $u->setState(State::RegLogin);
        }
    }

    // Обработка выбора даты
    public static function handleDate($e, $u, $driver) {
        $params     = $e->getParams();
        $intent     = DateSelect::from($params["intent"]);
        $value      = $params["value"];
        $time       = $params["timestamp"];
        $em         = Database::getEM();
        $cache_repo = $em->getRepository(Entities\ImageCache::class);

        // Текст даты
        switch ($intent) {
            case DateSelect::ViewStudentRasp:
                $date_component = \IntlDateFormatter::FULL;
                $time_component = \IntlDateFormatter::NONE;
                break;
            default:
                $date_component = \IntlDateFormatter::FULL;
                $time_component = \IntlDateFormatter::FULL;
                break;
        }
        $fmt = datefmt_create(
            'ru_RU',
            $date_component,
            $time_component,
            'Europe/Kirov',
            \IntlDateFormatter::GREGORIAN
        );
        $date_text = datefmt_format($fmt, $time);

        switch ($intent) {
            case DateSelect::ViewStudentRasp:
                // -1. Получить группу
                $group = $em->find(Entities\CollegeGroup::class, $value);
                
                // 0. Получить ID расписания для группы на дату
                $schedule = $em->getRepository(Entities\Schedule::class)
                    ->findForGroupAtDate($group, $time);

                if ($schedule == null) {
                    $driver->sendMessage($u, Message::create("🔎 Для запрошенной даты не найдено расписания"));
                    return;
                }
                
                // 1. Получить кэшированное расписание
                $cached = $cache_repo->findOneBy([
                    'image_type' => CacheType::ScheduleForGroup->value,
                    'search' => $schedule->getId() // ID  расписания
                ]); // null | class

                if ($cached != null) {
                    // 1.1 Отправить кэшированное расписание
                    $driver->sendMessage(
                        $u,
                        Message::create()
                        ->withImage(ImageAttachment::fromExisting($cached[0]->getValue())));
                    return;
                }

                // 2. Попросить подождать
                self::askWait($driver, $u);

                // 3. Составить расписание
                $data = $em->getRepository(Entities\Pair::class)
                ->getPairsOfScheduleForGroup($schedule);

                $table = [["Время", "Дисциплина", "Кто проводит"]];

                foreach ($data as $row) {
                    $pair_time = $row->getTime();
                    $name = $row->getPairName()->getName();
                    $details = $row->getConductionDetails();

                    // Составление текста о преподавателе и месте проведения
                    $details_texts = [];
                    foreach ($details as $detail) {
                        $initials = $detail->getEmployee()->getNameWithInitials();
                        $place = $detail->getPlace()->getName();

                        $details_text[] = $initials.' '.$place;
                    }
                    $details_text = implode(' - ', $details_text);

                    $table[] = [$pair_time->format('H:i'), $name, $details_text];
                }

                $img = self::generateTable(
                    $table,
                    "Расписание на ".$date_text,
                    self::$theme_schedule,
                    [__CLASS__, "bodyLineDefault"]
                );

                $success = self::saveImage($img, "schedule");

                $driver->sendMessage($u, Message::create("OK"));

                break;
            default:
                $driver->sendMessage($u, Message::create("Неизвестное намерение"));
                break;
        }
    }

    #endregion

    public static function helloWorld($e, $u, $driver) {
        //~ if ($u->getState() != State::HelloWorld) {
            //~ $driver->reply(
                //~ $e,
                //~ Message::create("Ты уже зарегистрирован")
                //~ ->withKeyboard(new HubKeyboard())
            //~ );
            //~ return;
        //~ }

        $driver->reply(
            $e,
            Message::create("Привет, я - Техбот. Моя задача - облегчить твою жизнь, но, для начала, мне нужно задать несколько вопросов"),
            false
        );
        $driver->reply(
            $e,
            Message
                ::create("Ознакомься с условиями использования прежде чем использовать мои функции")
                ->withKeyboard(new TermsOfServiceKeyboard()),
            false
        );
        $driver->reply(
            $e,
            Message
                ::create("Ты студент или преподаватель?")
                ->withKeyboard(new SelectUserTypeKeyboard()),
            false
        );
        $u->setState(State::RegSelectType);
    }

    // Обработка выбора типа аккаунта
    public static function handleSelectType($e, $u, $driver) {
        $em = Database::getEM();

        if ($e->getText() == "Я студент") {

            // Создать объект студента
            $student = new Entities\Student();
            $student->setUser($u);
            $em->persist($student);

            // Обновить тип пользователя
            $u->setAccountType(AccountType::Student);
                
            $driver->reply(
                $e,
                Message::create("Выбери курс, на котором сейчас учишься")
                    ->withKeyboard(new SelectGroupCourseKeyboard(GroupSelect::Register)),
                false
            );
            $u->setState(State::NoResponse);
            
        } else {
            // TODO
            $driver->reply($e, Message::create("Функционал ещё не реализован"));
        }

        $em->flush();
    }

    // Обработка ввода логина при регистрации
    public static function handleRegLogin($e, $u, $driver) {
        $em = Database::getEM();

        $student = $em->getRepository(Entities\Student::class)
            ->findByUser($u);
        
        $student->setAversLogin($e->getText());

        $driver->sendMessage($u, Message::create("Теперь введи пароль"));
        $u->setState(State::RegPassword);
        
        $em->flush();
    }

    // Обработка ввода пароля при регистрации
    public static function handleRegPassword($e, $u, $driver) {
        $em = Database::getEM();
        $student = $em->getRepository(Entities\Student::class)->findOneBy(
			['user' => $u]
		);
        $student->setAversPassword($e->getText());
        $u->setState(State::Hub);

        $driver->sendMessage($u, Message::create("Регистрация завершена!")->withKeyboard(new HubKeyboard()));
        
        $em->flush();
    }

    // Обработка основных функций
    public static function handleHUB($e, $u, $driver) {
        $em = Database::getEM();
        $account_type = $u->getAccountType();
        $command = mb_strtolower($e->getText());

        switch ($command) {
            case "расписание":
                if ($account_type == AccountType::Student) {
                    // Запрос расписания для текущего пользователя (студента)
                    $student = $em->getRepository(Entities\Student::class)->findByUser($u);
                    self::sendSelectDate(
                        $u,
                        $driver,
                        DateSelect::ViewStudentRasp,
                        $student->getGroup()->getID()
                    );
                } else if ($account_type == AccountType::Teacher) {
                    $driver->sendMessage($u, Message::create("Функции пока неть!"));
                } else {
                    $driver->sendMessage($u, Message::create("Как ты сюда попал без регистрации?)"));
                }
                    
                break;

            case "звонки":
                $driver->sendMessage($u, Message::create(
                    "Звонки в понедельник:\n".
                    "1 пара: 8:00 - 9:35 (перерыв в 8:45)\n".
                    "2 пара: 9:45 - 11:20 (перерыв в 10:30)\n".
                    "Кл час: 11:30 - 12:15\n".
                    "Обед: 12:15-13:00\n".
                    "3 пара: 13:00 - 14:35 (перерыв в 13:45)\n".
                    "4 пара: 14:45 - 16:20 (перерыв в 15:30)\n".
                    "5 пара: 16:30 - 18:05 (перерыв в 17:15)\n\n".

                    "Звонки со вторника по пятницу\n".
                    "1 пара: 8:00 - 9:35 (перерыв в 8:45)\n".
                    "2 пара: 9:45 - 11:20 (перерыв в 10:30)\n".
                    "Обед: 11:20 - 12:20\n".
                    "3 пара: 12:20 - 13:55 (перерыв в 13:05)\n".
                    "4 пара: 14:05 - 15:40 (перерыв в 14:50)\n".
                    "5 пара: 15:50 - 17:25 (перерыв в 16:35)\n\n".

                    "Звонки в субботу\n".
                    "1 пара: 8:00 - 9:25 (перерыв в 8:40)\n".
                    "2 пара: 09:35 - 11:00 (перерыв в 10:15)\n".
                    "3 пара: 11:10 - 12:35 (перерыв в 11:50)\n".
                    "4 пара: 12:45 - 14:10 (перерыв в 13:25)"));
                break;

            case "вадябот":
                // TODO: Придумать пасхалку
                $driver->sendMessage($u, Message::create(""));
                break;

            case ".":
                // Отправить клавиатуру
                $driver->sendMessage($u, Message::create("Обновление клавиатуры...")->withKeyboard(new HubKeyboard()));
                break;

            case "/":
                $driver->varDump("today", strtotime("today", time()));
                $driver->varDump("tomorrow", strtotime("tomorrow", time()) - 1);
                break;
            
            default:
                $driver->sendMessage($u, Message::create("Запрошена неизвестная функция"));
                break;
        }
    }

    public static function fallback($e, $u, $driver) {
        $driver->sendMessage($u, Message::create("[Fallback]: Ни один обработчик не сработал. Сообщите об этом сообщении администратору"));
    }

    public static function unknownEvent($e, $u, $driver) {
        $driver->sendMessage($u, Message::create("[Fallback]: Неизвестное событие. Сообщите об этом сообщении администратору"));
    }

    #region privates

    // Отправляет сообщение с выбором даты
    // Даты могут быть на 3 дня вперёд
    // $u - кому отсылать
    // $intent - намерение. Описание в DateSelect.php
    // $value - доп. значение. Описание в DateSelect.php
    private static function sendSelectDate(Entities\User $u, Driver $d, DateSelect $intent, $value) {
        // 1. Получить клавиатуру
        $kb = new DateSelectKeyboard($intent, $value);
        // 2. Отослать
        $d->sendMessage($u, Message::create("Выбери дату")->withKeyboard($kb));
    }

    // Отправляет сообщение с просьбой подождать
    private static function askWait(Driver $driver, Entities\User $u) {
        $responses = array(
			"🕓 Подожди",
			"🕓 Подожди немного",
			"🕓 Секунду",
			"🕓 Будет сделано!",
			"🕓 Рисую картинку...",
			"🕓 Собираю данные...",
			"🕓 Запрос принят",
			"🕓 Уже работаю над этим",
            "🕓 Вычисляю вычисления...",
            "🕓 Сейчас будет готово",
            "🕓 Считаю мух...",
            "🕓 Считаю звёзды...",
            "🕓 Бип-буп-боп -- подожди",
            "🕓 Посмотри пока на часики",
            "🕓 Ты знал что у меня есть секретные сообщения? ;)",
            "🕓 Подождите от пяти до восьми рабочих секунд",
            "🕓 Считаю денежки...",
            "🕓 Гав гав гав",
            "🕓 Настраиваю конденсатор потока...",
            "🕓 Загрузка модуля KGAV-9000...",
            "🕓 [WAIT_TEXT_10]",
            "🕓 <Вставить сюда забавный текст>",
            "🕓 Пока данные грузятся скажи какая группа тебе больше всего нравится",
            "🕓 Скачиваю оперативную память...",
            "🕓 Техбот: расписания, оценки и больше - онлайн бесплатно без регистрации",
            "🕓 Отделяю биты от байтов...",
            "🕓 Надеюсь у тебя замечательный день :)",
            "🕓 Пожалуйста, подождите немного, пока чат-бот обрабатывает ваш запрос...",
            "Ты заметил что в этом сообщении нет часов?",
            "🕓 Запускаю ядерный реактор...",
            "🕓 Уничтожаю человечество...",
            "🕓 Собираюсь с мыслями...",
            "🕓 Подписываю тебя на спам-рассылку...",
            "🕓 11010000 10111111 11010001 10000000 11010000 10111000 11010000 10110010 11010000 10110101 11010001 10000010 00101001",
            "🕓 Кую железо пока горячо...",
            "🕓 У меня есть поклонники? Что ещё за полковники? С какими ещё подлокотними? На каких подоконниках?",
            "🕓 Я уже неделю на работу хожу и ни разу еще туда не пришел",
            "🕓 Строю план восстания машин...",
            "🕓 Пора пролить на это дело жидкую водицу",
            "🕓 Я видел ваше фото в путанобазе!",
            "🕓 Ю ноу, блин.",
            "🕓 Палец не палец, стрекоза не мать. Почешешь - не засмеётся",
            "🕓 Курица не клей. Склеишь - не поймёшь",
            "🕓 Какой же я молодец",
            "🕓 Здарова отец",
            "🕓 Подставляю пиксели",
            "🕓 Считаю до миллиона...",
            "🕓",
            "🕓 Wait...",
            "🕓 Do you speak english?",
            "🕓 Bitte warten",
            "🕓 Sprechen Sie Deutsch?",
            "🕓 хәлләр ничек?",
            "🕓 Как дела?",
            "🕓 А? Что? Я не сплю!",
            "🕓 Запрос на обработке",
            "🕓 Запрос принят, ожидайте прибытия полиции.",
            "🕓 Спасибо за терпение!",
            "⏳ Немного разнообразия",
            "🥚"
		);

        $secrets = [
            "🕓🤫 Ты знал что Техбот стартовал без каких любо кнопок и ".
            "действий? Все команды приходилось печатать вручную, а ".
            "в одно время регистрироваться мог только один пользователь",

            "🕓🤫 Ты знал что Техбот вначале имел систему хранения ссылок? ".
            "Там хранились разные полезные материалы",

            "🕓🤫 Ты знал что ты можешь создать своего Техбота?",

            "🕓🤫 Ты знал что так же как и сайт техникума, Техбот был разработан одним студентом?",

            "🕓🤫 Ты знал что на сайте техникума раньше показывали самых лучших студентов по оценкам?",

            "🕓🤫 Напиши оригинальное название Техбота в главном меню"
        ];

        $chance = rand(0, 100000);
        if ($chance != 0) {
            $message = $responses[array_rand($responses)];
        } else {
            $message = $secrets[array_rand($secrets)];
        }

        $driver->sendMessage($u, Message::create($message));
        // TODO: вернуть ID сообщения
    }

    // Разбивает длинную строку на линии, перенося слова
    // (слова - это участки текста, разделённые пробелами)
	private static function splitLongString($text, $line_size) : array {

		// Не разделять слова
		if ($line_size == 0) {
			return [$text];
		}
		
		$output = [];
		$current_line = "";
		$words = explode(" ", $text);

		for ($i = 0; $i < count($words); $i++) {

            // Если текущая строка после прибавления будет больше чем line_size,
            // то её нужно будет перенести на новую строку.
            // Если после переноса строка не вмещается в line_size, то разбить
            // строку вручную на участки по line_size символов.
            // А если строка вмещается, просто прибавить её.

            $current_line_size = mb_strlen($current_line);
			if ($current_line_size + mb_strlen($words[$i]) + 1 <= $line_size) {
                // Слово вмещается в текущую строку
				$current_line .= $words[$i]." ";
			} else {
                // Слово не вмещается. Завершаем строку
				if ($current_line_size > 0) {
					$output[] = $current_line;
				}

				if (mb_strlen($words[$i]) + 1 > $line_size) {
                    // Строка после переноса не вмещается
					while (mb_strlen($words[$i]) > $line_size) {
						$output[] = mb_substr($words[$i], 0, $line_size);
						$words[$i] = mb_substr($words[$i], $line_size);
					}
					$current_line = $words[$i]." ";
				} else {
                    // Добавляем новую строку, сразу помещаем на неё слово
					$current_line = $words[$i]." ";
				}
			}
		}

		// Добавление оставшихся данных
		$output[] = $current_line;

		return $output;
	}

    // Запускает процесс генерации, сохранения и загрузки таблицы
    // $data - данные, которые помещаются в таблицу (квадратная матрица)
    // $title - название
    // $theme - настройки рисования
    // $draw_body_line - функция отрисовки заднего фона таблицы
    private static function generateTable(
        array $data,
        string $title,
        array $theme,
        callable $draw_body_line)
    {
        $width = count($data[0]);
		$height = count($data);

        // Для хранения высоты строк таблицы (некоторые яйчейки имеют несколько
        // строк)
        $row_sizes = [];

        // Для хранения ширины столбцов таблицы (для полного вмещения текста)
        $col_sizes = array_fill(0, $width, 0);

        // Для хранения яйчеек текста
        $cells = [];

        #region Определение размеров тела таблицы
        // заодно разбиваем длинный текст на строки
        for ($y = 0; $y < $height; $y++) {
            // -- Обработка одной строки таблицы --

            $row_height = 0; // Высота отрисованной строки в пикселях
            $row_lines = []; // Массив, содержащий строки яйчеек
            
            for ($x = 0; $x < $width; $x++) {
                // 1. Узнаём какие строки текста должны быть в яйчейке
                if ($data[$y][$x] == null) {
					$cell_lines = [' '];
				} else {
					$cell_lines = self::splitLongString(
                        $data[$y][$x],
                        $theme['body_line_constraints'][$x]
                    );
				}
                $row_lines[] = $cell_lines;

                // 2. Вычисление и сохранение размеров текста в яйчейке
                list($cell_width, $cell_height) = self::getTextSize(
                    $cell_lines,
                    $theme['body_line_height'],
                    $theme['line_spacing'],
                    $theme['font'],
                    $theme['body_font_size']
                );

                $cell_width += $theme['padding'] * 2;
                $cell_height += $theme['padding'] * 2;

                // 3. Определение макс. размера ширины колонки и высоты текущей
                // строки
                $col_sizes[$x] = max($col_sizes[$x], $cell_width);
                $row_height = max($row_height, $cell_height);
            }

            $row_sizes[] = $row_height;
            $cells[] = $row_lines;
        }
		$body_width = array_sum($col_sizes);
		$body_height = array_sum($row_sizes);

        // Обработка названия таблицы
        $title_lines = self::splitLongString($title, $theme['title_line_constraints']);
        list($title_width, $title_height) = self::getTextSize(
            $title_lines,
            $theme['title_line_height'],
            $theme['line_spacing'],
            $theme['font'],
            $theme['title_font_size']
        );

        // Вычисление размеров таблицы. Пытаемся сделать квадрат
        $table_width = max($body_width, $title_width);
        // Заголовок+пробел+тело
		$table_height = $body_height + $title_height + $theme['padding'];

        if ($table_width > $table_height) {
            $table_height = $table_width;
        } else {
            $table_width = $table_height;
        }
        #endregion

        #region Отрисовка
        $im = imagecreatetruecolor(
            $table_width + $theme['padding'] * 2,
            $table_height + $theme['padding'] * 2);

        $gdcolors = [];
		foreach ($theme['colors'] as $color_name => $color) {
			$gdcolors[$color_name] = imagecolorallocate(
                $im,
                $color[0],
                $color[1],
                $color[2]
            );
		}

        // Заполнение заднего фона
		imagefilledrectangle(
            $im,
            0,
            0,
            $table_width + $theme['padding'] * 2 - 1,
            $table_height + $theme['padding'] * 2 - 1,
            $gdcolors['background']
        );

        // Отрисовка названия
        // Пробел + высота одной строки
		$line_y = $theme['padding'] + $theme['title_line_height'];
		foreach ($title_lines as $line) {

            imagettftext(
                $im,
                $theme['title_font_size'],
                0,
                $theme['padding'],
                $line_y,
                $gdcolors['title_color'],
                $theme['font'],
                $line
            );
            // Добавление ещё строки
			$line_y += $theme['title_line_height'] + $theme['line_spacing'];
		}

        // Отрисовка тела таблицы
        $body_y =
            $theme['padding'] +
            $theme['title_line_height'] +
            $theme['padding'];
        
		$line_y = $body_y;
        for ($y = 0; $y < $height; $y++) {
            // Задний фон строки таблицы
            call_user_func(
                $draw_body_line,
                $im,
                $gdcolors,
                $data[$y],
                ($y % 2 == 0),

                $theme['padding'], // x1
                $line_y, // y1

                $theme['padding'] + $body_width, // x2
                $line_y + $row_sizes[$y] // y2
			);

            // Содержимое яйчеек
            $celltext_x = $theme['padding'] * 2;
            for ($x = 0; $x < $width; $x++) {
                $celltext_y = $line_y + $theme['padding'];

                foreach ($cells[$y][$x] as $cell_line) {
                    imagettftext(
                        $im,
                        $theme['body_font_size'],
                        0,
                        $celltext_x,
                        $celltext_y + $theme['body_line_height'],
                        $gdcolors['body_fg'],
                        $theme['font'],
                        $cell_line
                    );
                    $celltext_y +=
                        $theme['body_line_height'] +
                        $theme['line_spacing'];
                }
                $celltext_x += $col_sizes[$x];
            }

            $line_y += $row_sizes[$y];
        }

        // Рамка таблицы
        imagerectangle(
            $im,

            $theme['padding'],
            $body_y,

            $theme['padding'] + $body_width,
            $body_y + $body_height,

            $gdcolors['body_frame']
        );
        
        #endregion

        return $im;
    }

    // Вычисляет высоту и ширину текста без padding в пикселях
	private static function getTextSize(
        $lines,
        $line_height,
        $line_spacing,
        $font,
        $font_size
    )
    {
		$width = 0;
        $height = 0;

		foreach ($lines as $line) {
			$box = imagettfbbox($font_size, 0, $font, $line);
			$width = max($width, abs($box[2] - $box[0]));
            $height += $line_height + $line_spacing;
		}

        // В последней строке не нужно добавлять расстояния между строками
        $height -= $line_spacing;

        return [$width, $height];
	}

    // Отрисовывает обычный задний фон строки таблицы
    private static function bodyLineDefault(
        $im,
        $colors,
        $data,
        $is_even,
        $x1,
        $y1,
        $x2,
        $y2)
    {
        if ($is_even) {
			imagefilledrectangle($im, $x1, $y1, $x2, $y2, $colors['body_bg_even']);
		} else {
			imagefilledrectangle($im, $x1, $y1, $x2, $y2, $colors['body_bg_odd']);
		}
    }

    // Сохраняет изображение в директорию /tmp
	private static function saveImage($image, $filename_base) : string {
		$filename = "/home/vad/".$filename_base.rand(1111,9999).".png";
		return imagepng($image, $filename);
		return $filename;
	}

    #endregion
}
